/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Kamen Lisp.
 *
 * The Initial Developer of the Original Code is
 * Dustin Long (dlong@stevens.edu).
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifdef _WIN32
#include <windows.h>
#endif

#include <stdlib.h>
#include <string.h>

#ifndef _WIN32
#define stricmp strcasecmp
#endif

#include "ecl/ecl.h"
#include "../../common/marshalling.h"
#include "../../common/constants.h"

#ifdef _WIN32
BOOL APIENTRY DllMain(HANDLE hModule,
                      DWORD  ul_reason_for_call,
                      LPVOID lpReserved)
{
    switch (ul_reason_for_call)
    {
        case DLL_PROCESS_ATTACH:
        case DLL_THREAD_ATTACH:
        case DLL_THREAD_DETACH:
        case DLL_PROCESS_DETACH:
            break;
    }
    return TRUE;
}
#endif


static int s_isInitialized = 0;
static int s_securityVersion = 0;
static staticTable * s_staticTable = 0;
static void * s_domContext = 0;
cl_object s_finalizer = Cnil;

void eclRet(cl_object clIn, int isErr, ResultProc callback, void * context);
void eclDisp(cl_object clIn);
cl_object eclRun(const char * str);
cl_object eclExec(cl_object clCodeString);

cl_object internal_alert(cl_object sym);

cl_object internal_dom(cl_narg narg, ...);
cl_object internal_dom_set(cl_narg narg, ...);
cl_object internal_dom_get(cl_narg narg, ...);
cl_object internal_dom_call(cl_narg narg, ...);

cl_object finalize_dom_element(cl_object dom);


cl_object doQuote(cl_object obj)
{
    static cl_object quote = Cnil;
    if (quote == Cnil)
    {
        quote = cl_find_symbol(1, make_simple_base_string("QUOTE"));
    }
    return CONS(quote, CONS(obj, Cnil));
}

cl_object safeRead(cl_object stream, cl_object eopSym, int codeType)
{
    cl_object readSym = cl_find_symbol(1, make_simple_base_string("READ"));
    cl_object form = CONS(readSym,
                          CONS(doQuote(stream),
                               CONS(Cnil,
                                    CONS(doQuote(eopSym),
                                         Cnil))));
    if (codeType == 200)
    {
        cl_object got = si_safe_eval(3, form, Cnil, Cnil);
        return got;
    }
    else
    {
        cl_object got = ext_secure_safe_eval(3, form, Cnil, Cnil);
        return got;
    }
}


extern "C" int _stdcall evaluateLisp(const char * lisp, int codeType,
                                     ResultProc callback, void * context)
{
    if (!s_isInitialized)
    {
        const int argc = 1;
        char * argv[1];
        argv[0] = "henshin.dll";
        cl_boot_options(argc, argv, 0);

#ifdef _WIN32
        // TODO use grey streams
        eclRun("(setq *error-output* (make-string-output-stream))");
        eclRun("(setq *terminal-io* (make-string-output-stream))");
#endif

        eclRun("(make-package 'kamen-common)");
        eclRun("(in-package kamen-common)");
        eclRun("(setf document 'document window 'window)");

        eclRun("(defsetf dom dom-set)");
        eclRun("(export '(document window alert dom "
                         "dom-set dom-get dom-call "
                         DOM_OBJECT_NAME "))");

        cl_def_c_function(cl_intern(1, make_simple_base_string("ALERT")),
                          (void*)internal_alert, 1 );
        cl_def_c_function_va(cl_intern(1, make_simple_base_string("DOM")),
                             (void*)internal_dom);
        cl_def_c_function_va(cl_intern(1, make_simple_base_string("DOM-SET")),
                             (void*)internal_dom_set);
        cl_def_c_function_va(cl_intern(1, make_simple_base_string("DOM-GET")),
                             (void*)internal_dom_get);
        cl_def_c_function_va(cl_intern(1, make_simple_base_string("DOM-CALL")),
                             (void*)internal_dom_call);

        eclRun("(ffi:def-foreign-type " DOM_OBJECT_NAME " :pointer-void)");
        s_finalizer = cl_make_cfun((void*)finalize_dom_element, Cnil, NULL, 1);

        s_isInitialized = 1;
    }

    if (!s_securityVersion)
    {
        int s_securityVersion = security_version();
        if (s_securityVersion != 10101)
        {
            char * msg = "Preview version needs 1.01.01 "
                         "of ecl security system";
            cl_object readErr = make_simple_base_string(msg);
            eclRet(readErr, 1, callback, context);
            return 0;
        }        
    }


    int ret = 1;

    char * fstr = 0;

    if (codeType == 200)
    {
        fstr = "(list t ~s)";
    }
    else if (codeType == 1)
    {
        fstr = "(handler-case"
               "  `(t ,(ext:call-secure"
               "         (lambda () ~s)))"
               "  (error (e) `(nil ,e)))";
    }
    else if (codeType == 0)
    {
        // find function and run it
        cl_object funcSym = c_string_to_object((char*)lisp);
        cl_object isFunction = cl_fboundp(funcSym);
        if (isFunction != Cnil)
        {
            fstr = "(handler-case"
                   "  `(t ,(ext:call-secure #'~a))"
                   "  (error (e) `(nil ,e)))";
        }
        else
        {
            printf("not a function\n");
            return 0;
        }
    }
    else
    {
        printf("Unknown code type\n");
        return 0;
    }

    cl_object formatStr = make_simple_base_string(fstr);

    cl_object outc, actual, made = Cnil;
    int isErr = 0;

    cl_object text = make_simple_base_string((char*)lisp);
    cl_object stream = cl_make_string_input_stream(1, text);
    cl_object eopSym = cl_gensym(0);

    // read until we have finished with all forms
    for (;;)
    {
        outc = safeRead(stream, eopSym, codeType);

        if (outc == Cnil)
        {
            cl_object readErr = make_simple_base_string("Read error");
            made = CONS(Cnil, CONS(readErr, Cnil));
            isErr = 1;
            break;
        }

        if (&(outc->symbol) == &(eopSym->symbol))
        {
            break;
        }

        actual = cl_format(3,
                           Cnil,
                           formatStr,
                           outc);

        made = eclExec(actual);

        if (CAR(made) == Cnil)
        {
            isErr = 1;
            break;
        }

        if (codeType == 0)
        {
            break;
        }
    }
    // send the result to the callback
    eclRet(CAR(CDR(made)), isErr, callback, context);
    eclRun("(si::gc t)");

    return ret;
}

extern "C" int _stdcall setStaticTable(staticTable * table)
{
    s_staticTable = table;

    return 1;
}

cl_object eclRun(const char * str)
{
    cl_object form = c_string_to_object(str);
    cl_object result = si_safe_eval(3, form, Cnil, Cnil);
    return result;
}

cl_object eclRunUnsafe(const char * str)
{
    cl_object form = c_string_to_object(str);
    cl_object result = cl_eval(form);
    return result;
}

cl_object eclExec(cl_object clCodeString)
{
    cl_object form = cl_read_from_string(1, clCodeString);
    cl_object result = si_safe_eval(3, form, Cnil, Cnil);
    return result;
}

void eclRet(cl_object clIn, int isErr, ResultProc callback, void * context)
{
    cl_object show = cl_princ_to_string(1, clIn);
    char * got = (char *)show->base_string.self;
    callback(got, isErr, context);
}

void eclDisp(cl_object clIn)
{
    cl_object show = cl_write_to_string(1, clIn);
    char * got = (char *)show->base_string.self;
    printf("%s", got);
}

void addElement(variantHolder * target, cl_object obj);
void retrieveElement(cl_object * make, variantHolder source);

void executeAlertCall(cl_object obj)
{
    variantHolder local;
    variantHolderArray build;
    build.size = 1;
    build.array = &local;
    addElement(&local, obj);
    s_staticTable->alertCall(s_staticTable->self, build);
    destroyVariant(&local);
}

cl_object executeDomAction(int num, cl_object * objs, char * action, int * err)
{
    cl_object ret = Cnil;
    int size = num + 1;

    cl_object * each = (cl_object*)malloc(sizeof(cl_object) * size);
    each[0] = Cnil;

    variantHolder outc;
    variantHolderArray build;
    build.size = size;
    build.array = (variantHolder*)malloc(sizeof(variantHolder) * size);

    build.array[0] = makeSymbolVariant(action);

    int k;
    for (k = 1; k < size; k++)
    {
        each[k] = objs[k - 1];
        addElement(&build.array[k], each[k]);
    }

    s_staticTable->domAction(&outc, s_staticTable->self, build, err);

    int j;
    for (j = 0; j < size; j++)
    {
        destroyVariant(&build.array[k]);
    }

    retrieveElement(&ret, outc);

    free(build.array);
    free(each);

    destroyVariant(&outc);

    return ret;
}

void signalIfError(int err)
{
    if (err == DOM_ERR_LACK)
    {
        eclRunUnsafe("(error \"Too few arguments to 'dom'\")");
    }
    else if (err == DOM_ERR_EXTRA)
    {
        eclRunUnsafe("(error \"Too many arguments to 'dom'\")");
    }
    else if (err == DOM_ERR_FATAL)
    {
        eclRunUnsafe("(error \"Fatal error\")");
    }
    else if (err == DOM_ERR_FIELD)
    {
        eclRunUnsafe("(error \"Field is undefined\")");
    }
    else if (err == DOM_ERR_UNCALL)
    {
        eclRunUnsafe("(error \"Not a callable function\")");
    }
    else if (err != 0)
    {
        eclRunUnsafe("(error \"Unknown error\")");
    }
}


cl_object internal_dom(cl_narg narg, ...)
{
    cl_object ret;
    int err = 0;
    cl_va_list params;
    cl_va_start(params, narg, narg, 0);
    cl_object * args = (cl_object*)malloc(sizeof(cl_object) * narg);
    int k;
    for (k = 0; k < narg; k++)
    {
        args[k] = cl_va_arg(params);
    }
    ret = executeDomAction(narg, args, "DOM", &err);

    free(args);

    signalIfError(err);

    //cl_princ(1, ret);

    return ret;
}

cl_object internal_dom_set(cl_narg narg, ...)
{
    cl_object ret;
    int err = 0;
    cl_va_list params;
    cl_va_start(params, narg, narg, 0);
    cl_object * args = (cl_object*)malloc(sizeof(cl_object) * narg);
    int k;
    for (k = 0; k < narg; k++)
    {
        args[k] = cl_va_arg(params);
    }
    ret = executeDomAction(narg, args, "DOM-SET", &err);
    free(args);
    signalIfError(err);
    return ret;
}

cl_object internal_dom_get(cl_narg narg, ...)
{
    cl_object ret;
    int err = 0;
    cl_va_list params;
    cl_va_start(params, narg, narg, 0);
    cl_object * args = (cl_object*)malloc(sizeof(cl_object) * narg);
    int k;
    for (k = 0; k < narg; k++)
    {
        args[k] = cl_va_arg(params);
    }
    ret = executeDomAction(narg, args, "DOM-GET", &err);
    free(args);
    signalIfError(err);
    return ret;
}

cl_object internal_dom_call(cl_narg narg, ...)
{
    cl_object ret;
    int err = 0;
    cl_va_list params;
    cl_va_start(params, narg, narg, 0);
    cl_object * args = (cl_object*)malloc(sizeof(cl_object) * narg);
    int k;
    for (k = 0; k < narg; k++)
    {
        args[k] = cl_va_arg(params);
    }
    ret = executeDomAction(narg, args, "DOM-CALL", &err);
    free(args);
    signalIfError(err);
    return ret;
}


cl_object internal_alert(cl_object sym)
{
    cl_object show = cl_princ_to_string(1, sym);
    executeAlertCall(show);
    return show;
}

cl_object finalize_dom_element(cl_object dom)
{
    if (type_of(dom) == t_foreign)
    {
        ecl_foreign * f = &(dom->foreign);
        char * foreign_tag = (char*)f->tag->symbol.name->base_string.self;
        if (stricmp(foreign_tag, DOM_OBJECT_NAME) == 0)
        {
            void * pointer = f->data;
            s_staticTable->finalizeDom(s_staticTable->self, pointer);
            si_free_foreign_data(dom);
        }
    }

    return Cnil;
}

void addElement(variantHolder * target, cl_object obj)
{
    *target = variantHolder();
    if (type_of(obj) == t_symbol)
    {
        *target = makeSymbolVariant((char*)obj->symbol.name->base_string.self);
    }
    else if (type_of(obj) == t_base_string)
    {
        char * string_contents = (char*)obj->base_string.self;
        if (obj->base_string.hasfillp)
        {
            cl_index leng = obj->base_string.fillp;
            if (string_contents[leng] != '\0')
            {
                // TODO debug message about realloc
                char * alloc_contents = (char*)malloc(leng + 1);
                memcpy(alloc_contents, string_contents, leng);
                alloc_contents[leng] = '\0';
                *target = makeAllocatedVariant(alloc_contents);
                return;
            }
        }
        *target = makeTextVariant(string_contents);
    }
    else if (type_of(obj) == t_fixnum)
    {
        int num = fix(obj);
        *target = makeIntVariant(num);
    }
    else if (type_of(obj) == t_foreign)
    {
        ecl_foreign * f = &(obj->foreign);
        char * foreign_tag = (char*)f->tag->symbol.name->base_string.self;
        if (stricmp(foreign_tag, DOM_OBJECT_NAME) == 0)
        {
            *target = makeDomVariant(f->data);
        }
        else
        {
            printf("Bad foreign data found!\n");
        }
    }
    else
    {
        printf("!!!unknown type [%x] [%x]!!", type_of(obj), obj);
        cl_object str = cl_write_to_string(1, obj);
        printf(" {%s}\n", str->base_string.self);
    }
}

void retrieveElement(cl_object * make, variantHolder source)
{
    char * text;
    int num;
    void * data;
    if (text = fetchVariantSymbol(source))
    {
        *make = cl_intern(1, make_simple_base_string(text));
    }
    else if (text = fetchVariantText(source))
    {
        *make = cl_copy_seq(make_simple_base_string(text));
    }
    else if (fetchVariantInt(source, &num))
    {
        *make = MAKE_FIXNUM(num);
    }
    else if (data = fetchVariantDom(source))
    {
        cl_object obj;
        obj = eclRun("(ffi::allocate-foreign-object '" DOM_OBJECT_NAME ")");
        obj->foreign.data = (char*)data;
        si_set_finalizer(obj, s_finalizer);
        *make = obj;
    }
    else
    {
        // mostly likely void type
        *make = Cnil;
    }
    // TODO debug message about data creation
}

