/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Kamen Lisp.
 *
 * The Initial Developer of the Original Code is
 * Dustin Long (dlong@stevens.edu).
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef SARUNOKUMO_CPP
#define SARUNOKUMO_CPP


#ifdef _WIN32
#include <windows.h>
#endif

#include "SaruNoKumo.h"
#include "../../common/constants.h"

#include "nsCOMPtr.h"
#include "nsIFile.h"
#include "nsStringAPI.h"
#include "nsEmbedString.h"

#include "nsIDOMWindow.h"

#include "nsIExtensionManager.h"
#include "nsServiceManagerUtils.h"

#include "nsComponentManagerUtils.h"



#include "jsapi.h"

#include "nsIJSContextStack.h"

#include "nsIXPConnect.h"

#include "nsIJSRuntimeService.h"

#include "nsIXPCScriptable.h"



#include "../../meta/src/js-bridge.cpp"

#include <map>
using std::map;

struct ltstr
{
    bool operator()(const char* s1, const char* s2) const
    {
        return strcmp(s1, s2) < 0;
    }
};

typedef std::map<const char *, char *, ltstr>::iterator cache_result;



class SaruNoKumoImpl
{
  public:

    void getXPConnect(nsIXPConnect ** got);

    void assignContext(nsIXPConnect * xpc);

    int convertArguments(variantHolderArray parms,
                         int * pargc, jsval * * pargv);

    int convertResult(nsIXPConnect * xpc, jsval val, variantHolder * ret);

    char * enumerateFields(JSObject * obj, char * property, variantHolder * stringable);

    int fieldMatch(char * want, char * real);

    void wrapSupports(nsIXPConnect * xpc, nsISupports * supports,
                      nsIXPConnectJSObjectHolder * * holder);

    int hostToJSHolder(nsIXPConnect * xpc, variantHolder host,
                       nsCOMPtr<nsIXPConnectJSObjectHolder> & jshold);

    char * properField(nsIXPConnectJSObjectHolder * jshold, variantHolder field,
                       variantHolder * propStringable);

    void executeJavascript(dom_action_mode mode,
                           nsIXPConnect * xpc, JSObject * obj,
                           char * property, int argc, jsval * argv,
                           jsval * outc, int * err);

    int symbolCapitalization(char * dest, char * sour);

    int findInCache(char * search, char ** found);

    void storeInCache(char * search, char * found);

  private:
    JSContext * m_cx;
    map<const char *, char *, ltstr> cache;

};


void SaruNoKumo::init()
{
    JS_bridgeInit();
    m_impl = new SaruNoKumoImpl();
}

variantHolder SaruNoKumo::domAction(dom_action_mode mode, variantHolder host,
                                    variantHolder field,
                                    variantHolderArray params, int * err)
{
    variantHolder outc;

    do
    {
        nsCOMPtr<nsIXPConnect> xpc = do_GetService("@mozilla.org/js/xpc/XPConnect;1");
        if (!xpc) break;

        m_impl->assignContext(xpc);

        nsCOMPtr<nsIXPConnectJSObjectHolder> jshold;
        if (!m_impl->hostToJSHolder(xpc, host, jshold)) break;

        variantHolder propStringable;
        char * property = m_impl->properField(jshold, field, &propStringable);
        if (!property)
        {
            *err = DOM_ERR_FIELD; break;
        }
        do
        {
            int argc;
            jsval * argv;

            if (!m_impl->convertArguments(params, &argc, &argv)) break;

            JSObject * obj = 0;
            jshold->GetJSObject(&obj);
            if (!obj) break;

            jsval got;
            m_impl->executeJavascript(mode, xpc, obj, property,
                                      argc, argv, &got, err);
            if (*err) break;

            m_impl->convertResult(xpc, got, &outc);
        }
        while (0);
        destroyVariant(&propStringable);

    }
    while (0);

    return outc;
}


int SaruNoKumoImpl::convertArguments(variantHolderArray parms,
                                     int * pargc, jsval * * pargv)
{
    int size = parms.size;

    jsval * args = (jsval*)malloc(size * sizeof(jsval));

    for (int k = 0; k < size; k++)
    {
        variantHolder & p = parms.array[k];

        variantHolder holder;

        // TODO assumes all parameters are strings

        char * data = variantStringable(&holder, p);
        int size = strlen(data);

        char * place = (char*)JS_bridge_malloc(m_cx, size + 1);
        memcpy(place, data, size + 1);

        if (fetchVariantSymbol(p))
        {
            size = this->symbolCapitalization(place, place);
        }

        // TODO (size + 1)?
        JSString * str = JS_bridge_NewString(m_cx, place, size);
        destroyVariant(&holder);

        args[k] = STRING_TO_JSVAL(str);
    }

    *pargc = size;
    *pargv = args;

    return 1;
}

int SaruNoKumoImpl::symbolCapitalization(char * dest, char * sour)
{
    char * made = dest;

    int len = 0;
    for (;;)
    {
        if (!(*sour)) { break; }

        if (sour[0] == '-')
        {
            dest[0] = toupper(sour[1]);
            sour++;
        }
        else
        {
            dest[0] = tolower(sour[0]);
        }
        dest++;
        sour++;
        len++;
    }
    *dest = 0;

    return len;
}

int SaruNoKumoImpl::convertResult(nsIXPConnect * xpc, jsval val, variantHolder * ret)
{
    if (JSVAL_IS_VOID(val))
    {
        return 0;
    }
    else if (JSVAL_IS_NULL(val))
    {
        return 0;
    }
    // what about integer?
    else if (JSVAL_IS_STRING(val))
    {
        JSString * str = JSVAL_TO_STRING(val);
        char * txt = JS_bridge_GetStringBytes(str);
        size_t len = JS_bridge_GetStringLength(str);

        char * make = (char*)malloc(len + 1);
        memcpy(make, txt, len);
        make[len] = '\0';

        *ret = makeAllocatedVariant(make);
        return 1;
    }
    else if (JSVAL_IS_BOOLEAN(val))
    {
        int value = JSVAL_TO_BOOLEAN(val);
        if (value)
        {
            *ret = makeTextVariant("true");
        }
        else
        {
            *ret = makeTextVariant("false");
        }
    }
    else if (JSVAL_IS_OBJECT(val))
    {
        do
        {
            JSObject * objp = 0;
            JS_bridge_ValueToObject(m_cx, val, &objp);

            if (!objp) break;

            nsCOMPtr<nsIXPConnectWrappedNative> wrapret;
            xpc->GetWrappedNativeOfJSObject(m_cx, objp, getter_AddRefs(wrapret));
            if (!wrapret) break;

            nsCOMPtr<nsISupports> supret;
            wrapret->GetNative(getter_AddRefs(supret));
            if (!supret) break;

            nsISupports * raw = supret;
            raw->AddRef();

            *ret = makeDomVariant(raw);
            return 1;
        }
        while (0);

        return 0;
    }
    else
    {
        printf("@@@@@@ unknown jsval - during GETTER @@@@@ %x\n", val);
        return 0;
    }
}


char * SaruNoKumoImpl::enumerateFields(JSObject * obj, char * want, variantHolder * stringable)
{
    JSIdArray * ids = JS_bridge_Enumerate(m_cx, obj);
    jsint leng = ids->length;

    for (jsint k = 0; k < leng; k++)
    {
        jsid i = ids->vector[k];
        jsval v;
        JS_bridge_IdToValue(m_cx, i, &v);
        JSString * s = JSVAL_TO_STRING(v);
        char * real = JS_bridge_GetStringBytes(s);

        if (this->fieldMatch(want, real))
        {
            int len = strlen(real);
            char * alloc = (char*)malloc(len + 1);
            memcpy(alloc, real, len);
            alloc[len] = 0;

            *stringable = makeAllocatedVariant(alloc);
            break;
        }
    }
    JS_bridge_DestroyIdArray(m_cx, ids);

    return fetchVariantText(*stringable);
}

int SaruNoKumoImpl::fieldMatch(char * want, char * real)
{
    for (;;)
    {
        if (!(*want) && !(*real))
        {
            return 1;
        }
        if (!(*want) || !(*real))
        {
            return 0;
        }
        if (*want == '-')
        {
            want++;
        }
        else if (tolower(*want) == tolower(*real))
        {
            want++;
            real++;
        }
        else
        {
            return 0;
        }
    }
}



void SaruNoKumoImpl::getXPConnect(nsIXPConnect ** got)
{
    nsCOMPtr<nsIXPConnect> xpc = do_GetService("@mozilla.org/js/xpc/XPConnect;1");
    if (!xpc) return;
    *got = xpc;
}


void SaruNoKumoImpl::assignContext(nsIXPConnect * xpc)
{
    nsresult rv;

    nsCOMPtr<nsIXPCNativeCallContext> call;
    rv = xpc->GetCurrentNativeCallContext(getter_AddRefs(call));

    if (call)
    {
        JSContext * cx;
        call->GetJSContext(&cx);
        m_cx = cx;
    }
}

int SaruNoKumoImpl::hostToJSHolder(nsIXPConnect * xpc, variantHolder host, nsCOMPtr<nsIXPConnectJSObjectHolder> & jshold)
{
    nsISupports * supports = (nsISupports*)fetchVariantDom(host);
    if (!supports) { return 0; }

    JSObject * scope = JS_bridge_GetScopeChain(m_cx);

    nsresult rv;
    rv = xpc->WrapNative(m_cx, scope, supports, NS_GET_IID(nsISupports), getter_AddRefs(jshold));

    if (!jshold) { return 0; }

    return 1;
}

char * SaruNoKumoImpl::properField(nsIXPConnectJSObjectHolder * jshold, variantHolder field, variantHolder * propStringable)
{
    // TODO make sure strings are kept the way they are

    char * real = 0;

    variantHolder holder;
    do
    {
        char * want = variantStringable(&holder, field);

        if (fetchVariantText(field))
        {
            real = want;

            break;
        }

        if (this->findInCache(want, &real))
        {
            // should check if that is actually valid for this object
            if (1)
            {
                break;
            }
            else
            {
                real = 0;
            }
        }

        do
        {
             {
                 JSObject * base = 0;
                 jshold->GetJSObject(&base);

                 if (!base) break;

                 real = this->enumerateFields(base, want, propStringable);

                 if (real) break; // SUCCESS
             }

             {
                 nsCOMPtr<nsIXPConnectWrappedNative> wrappedNative;
                 wrappedNative = do_QueryInterface(jshold);

                 if (!wrappedNative) break;

                 JSObject * proto = 0;
                 wrappedNative->GetJSObjectPrototype(&proto);

                 if (!proto) break;

                 real = this->enumerateFields(proto, want, propStringable);

                 if (real) break; // SUCCESS
             }
        }
        while (0);

        if (real)
        {
            this->storeInCache(want, real);
        }
    }
    while (0);
    destroyVariant(&holder);

    return real;
}

int SaruNoKumoImpl::findInCache(char * search, char ** found)
{
    cache_result result = cache.find(search);
    if (result != cache.end())
    {
        *found = (*result).second;
        return 1;
    }
    return 0;
}

void SaruNoKumoImpl::storeInCache(char * search, char * found)
{
    int len = strlen(found);
    char * make = (char*)malloc(len + 1);
    strcpy(make, found);
    cache[search] = make;
}

void SaruNoKumoImpl::wrapSupports(nsIXPConnect * xpc, nsISupports * supports, nsIXPConnectJSObjectHolder * * holder)
{
    JSObject * scope = JS_bridge_GetScopeChain(m_cx);
    nsCOMPtr<nsIXPConnectJSObjectHolder> jshold;
    nsresult rv;
    rv = xpc->WrapNative(m_cx, scope, supports, NS_GET_IID(nsISupports), getter_AddRefs(jshold));
    if (jshold)
    {
        *holder = jshold;
    }
    return;
}

void SaruNoKumoImpl::executeJavascript(dom_action_mode mode,
                                       nsIXPConnect * xpc, JSObject * obj,
                                       char * property, int argc, jsval * argv,
                                       jsval * outc, int * err)
{
    int ambiguousCall = 0;
    int callZeroArgFunc = 0;

    if (mode == DOM_ACTION_DEFAULT)
    {
        if (argc == 0)
        {
            ambiguousCall = 1;
            mode = DOM_ACTION_GET;
        }
        else
        {
            mode = DOM_ACTION_CALL;
        }
    }

    if (mode == DOM_ACTION_SET)
    {
        *outc = argv[0];

        JS_bridge_SetProperty(m_cx, obj, property, outc);
    }
    else if (mode == DOM_ACTION_GET)
    {
        JS_bridge_GetProperty(m_cx, obj, property, outc);

        if (ambiguousCall)
        {
            if (JSVAL_IS_OBJECT(*outc))
            {
                JSObject * made = 0;
                JS_bridge_ValueToObject(m_cx, *outc, &made);

                if (made &&JS_bridge_ObjectIsFunction(m_cx, made))
                {
                    callZeroArgFunc = 1;
                    argc = 0;
                    argv = 0;
                }
            }
        }
    }
    else if (mode == DOM_ACTION_CALL)
    {
        JS_bridge_GetProperty(m_cx, obj, property, outc);

        if (!JSVAL_IS_OBJECT(*outc))
        {
            *err = DOM_ERR_UNCALL;
            return;
        }

        JSObject * made = 0;
        JS_bridge_ValueToObject(m_cx, *outc, &made);
            
        if (!made || !JS_bridge_ObjectIsFunction(m_cx, made))
        {
            *err = DOM_ERR_UNCALL;
            return;
        }
    }

    if (mode == DOM_ACTION_CALL || callZeroArgFunc)
    {
        JS_bridge_CallFunctionName(m_cx, obj, property, argc, argv, outc);
    }
}



#endif
